<?php

/*
Plugin Name: Foundation UI
Author: Stefen Enns
Version: 1.0.0
*/

namespace Foundation {

    use ArrayAccess;
    use Foundation\Response\SessionResponse;
    use Foundation\Response\StoredResponse;
    use PandaRoseUI;

    abstract class Response implements ArrayAccess
    {
        protected $constructor;
        protected $type;
        protected $version;
        protected $page;

        public static function factory(...$args)
        {
            return new StoredResponse(...$args);
        }

        public static function version($version, $constructor = null, $type = null)
        {
            $class = get_called_class();

            return $class::factory($constructor, $type, $version);
        }

        public static function type($type, $constructor = null, $version = null)
        {
            $class = get_called_class();

            return $class::factory($constructor, $type, $version);
        }

        protected function initialize()
        {
        }

        protected function normalizeKey($key)
        {
            if ('__ui_' === substr($key, 0, 5)) {
                return $key;
            };

            if (substr($key, 0, 1) === '#') {
                $prefix = is_user_logged_in()
                    ? get_current_user_id()
                    : 'guest';
                $key = $prefix . '__' . trim($key, '#');
            }

            return '__ui_' . $key;
        }

        public function default($key, $value)
        {
            if (isset($this[$key])) {
                return $this[$key];
            }

            $this[$key] = $value;

            return $value;
        }

        public function set($key, $value)
        {
            $this[$key] = $value;

            return $value;
        }

        public function incr($key)
        {
            $val = $this[$key];

            if (!is_numeric($val)) {
                $val = 0;
            }

            $val++;

            $this->set($key, $val);

            return $val;
        }

        public function decr($key)
        {
            $val = $this[$key];

            if (!is_numeric($val)) {
                $val = 0;
            }

            $val--;

            if ($val < 0)
                $val = 0;

            $this->set($key, $val);

            return $val;
        }

        public function merge($key, array $value = [], $persist = true)
        {
            $val = $this[$key];

            if (!is_array($val)) {
                $val = [];
            }

            $val = array_merge($val, $value);

            if ($persist)
                $this->set($key, $val);

            return $val;
        }

        final public function __construct(
            $constructor = null,
            $type = null,
            $version = null,
            $page = null,
        ) {
            if (is_string($constructor) and !$type) {
                $type = $constructor;
                $constructor = null;
            }

            if (!$page and PandaRoseUI::$top)
                $page = PandaRoseUI::$top;

            $this->constructor = $constructor;
            $this->type = $type;
            if ($version === true)
                $this->version = true;
            else
                $this->version = $version ? (string) $version : null;
            $this->page = $page;

            $this->initialize();
        }

        protected function callConstructor()
        {
            if ($this->constructor and $_SERVER['REQUEST_METHOD'] === "GET") {
                $constructor = $this->constructor;
                $constructor = $constructor->bindTo($this, $this);

                $constructor();
            }
        }
    }
}

namespace Foundation\Response {

    use ArrayObject;
    use Foundation\Response;
    use IteratorAggregate;
    use PandaRoseUI;
    use Traversable;

    class SessionResponse extends Response implements IteratorAggregate
    {

        public function getIterator(): Traversable
        {
            $result = new ArrayObject;

            $key = $this->getSessionKey();

            if (isset($_SESSION[$key]) and is_array($_SESSION[$key]))
                foreach ($_SESSION[$key] as $k => $v)
                    $result[$k] = $v;

            return $result;
        }

        #[\ReturnTypeWillChange]
        public function offsetExists(mixed $offset)
        {
            $offset = $this->normalizeKey($offset);

            return $this[$offset] !== null;
        }

        #[\ReturnTypeWillChange]
        public function offsetGet(mixed $offset)
        {
            $offset = $this->normalizeKey($offset);

            $key = $this->getSessionKey();

            return PandaRoseUI::keyStringGet($_SESSION[$key], $offset);
        }

        #[\ReturnTypeWillChange]
        public function offsetSet(mixed $offset, mixed $value)
        {
            $offset = $this->normalizeKey($offset);

            $key = $this->getSessionKey();

            PandaRoseUI::keyString($offset, $value, $_SESSION[$key]);
        }

        #[\ReturnTypeWillChange]
        public function offsetUnset(mixed $offset)
        {
            $offset = $this->normalizeKey($offset);

            $key = $this->getSessionKey();

            return PandaRoseUI::keyString($offset, null, $_SESSION[$key]);
        }

        protected function initialize()
        {
            if (!session_id()) session_start();

            $key = $this->getSessionKey();

            if (!isset($_SESSION[$key])) {
                $_SESSION[$key] = [];

                $this->callConstructor();
            }
        }

        public function getSessionKey()
        {
            if (!$this->page) return 'unnamed_session';

            return md5($this->page) . 'z';
        }

        public function save(): Response
        {
            return $this;
        }
    }

    class StoredResponse extends SessionResponse
    {
        protected $orig = [];
        protected $meta = [];
        protected $sess;
        protected $post;

        #[\ReturnTypeWillChange]
        public function offsetExists(mixed $offset)
        {
            $offset = $this->normalizeKey($offset);

            if ($this->post) {
                return PandaRoseUI::keyStringGet($this->meta, $offset) !== null;
            }

            return parent::offsetExists($offset);
        }

        #[\ReturnTypeWillChange]
        public function offsetGet(mixed $offset)
        {
            $offset = $this->normalizeKey($offset);

            if ($this->post) {
                return PandaRoseUI::keyStringGet($this->meta, $offset);
            }

            return parent::offsetGet($offset);
        }

        #[\ReturnTypeWillChange]
        public function offsetSet(mixed $offset, mixed $value)
        {
            $offset = $this->normalizeKey($offset);

            if ($this->post) {
                PandaRoseUI::keyString($offset, $value, $this->meta);
                list($k) = explode('.', $offset);
                $v = $this->meta[$k];
                if (is_object($v) or is_array($v))
                    $v = 'data:' . base64_encode(json_encode($v));
                else
                    $v = json_encode($v);
                update_post_meta($this->post->ID, $k, $v);
            } else {
                parent::offsetSet($offset, $value);
            }
        }

        #[\ReturnTypeWillChange]
        public function offsetUnset(mixed $offset)
        {
            $offset = $this->normalizeKey($offset);

            if ($this->post) {
                $this[$offset] = null;
            } else {
                parent::offsetUnset($offset);
            }
        }

        protected function initialize()
        {
            $init = false;

            if (is_admin()) {
                if ($this->post = get_post()) {
                    $this->loadMetaData();
                }
            } else {
                if (!is_user_logged_in())
                    if (!session_id())
                        session_start();

                $args = array(
                    'post_type' => 'response',
                    'orderby'   => 'date',
                    'order'     => 'DESC',
                    'post_status' => 'any',
                    'meta_query'  => [
                        [
                            'key'   => '__ui_owner',
                            'value' => is_user_logged_in()
                                ? get_current_user_id()
                                : session_id(),
                        ]
                    ]
                );

                $result = new \WP_Query($args);

                if (isset($result->posts[0])) {
                    if ($this->post = $result->posts[0]) {
                        $this->loadMetaData();
                    }
                } else {
                    if (!session_id()) session_start();

                    $this->sess = $this->getSessionKey();

                    if (!isset($_SESSION[$this->sess])) {
                        $_SESSION[$this->sess] = [];
                        $init = true;
                    }
                }

                if ($this->version === true) {
                    $init = true;
                    $this['__version'] = null;
                } else if ($this->version !== null) {
                    $version = $this['__version'];
                    if ($version !== $this->version) {
                        $init = true;
                        $this['__version'] = $this->version;
                    }
                }
            }

            if ($init) {
                $this->callConstructor();
            }
        }

        protected function loadMetaData()
        {
            if (is_array($meta = get_post_meta($this->post->ID))) {
                foreach ($meta as $k => $v) {
                    $v = trim(implode("\n", $v));
                    if ('data:' === substr($v, 0, 5)) {
                        $v = base64_decode(substr($v, 5));
                    }
                    $this->meta[$k] = json_decode($v, true);
                }
            }
        }

        public function save(): Response
        {
            if ($this->sess && !$this->post) {
                $content = '';
                if (isset($this['state'])) {
                    $content = json_encode($this['state']);
                }

                $data = array(
                    'post_title'   => '#' . time(),
                    'post_content' => $content,
                    'post_type'    => 'response',
                    'post_status'  => 'draft',
                    'meta_input'   => []
                );

                foreach ($this as $key => $val) {
                    if ($key == 'state') continue;
                    if (is_array($val) || is_object($val))
                        $val = 'data:' . base64_encode(json_encode($val));
                    else
                        $val = json_encode($val);
                    $key = $this->normalizeKey($key);
                    $data['meta_input'][$key] = $val;
                }

                $data['meta_input']['__ui_owner'] = 1;

                wp_insert_post($data);
            }

            return $this;
        }
    }
}

namespace {
    if (defined('PANDA_UI')) return;

    require_once(__DIR__ . '/vendor/autoload.php');

    define('PANDA_UI', '1.0.11');
    define('FOUNDATION_UI', PANDA_UI);

    use buzzingpixel\twigswitch\SwitchTwigExtension;
    use Symfony\Component\Yaml\Yaml;
    use ScssPhp\ScssPhp\Compiler;

    class PandaRoseUIProxy
    {
        protected $data = [];
        public $last = [];

        public function __get($key)
        {
            return isset($this->data[$key]) ? $this->data[$key] : null;
        }
        public function __set($key, $val)
        {
            $this->data[$key] = $val;
        }
        public function __isset($key)
        {
            return isset($this->data[$key]);
        }
    }

    class PandaRoseData implements ArrayAccess
    {
        protected $data = [];

        public function __construct($data = [])
        {
            foreach ($data as $k => $v) {
                if (is_numeric($k)) {
                    $this->data[$v] = true;
                } else {
                    $this->data[$k] = $v;
                }
            }
        }

        #[\ReturnTypeWillChange]
        public function offsetExists($offset)
        {
            return isset($this->data[$offset]);
        }

        #[\ReturnTypeWillChange]
        public function offsetGet($offset)
        {
            return isset($this->data[$offset]) ? $this->data[$offset] : null;
        }

        #[\ReturnTypeWillChange]
        public function offsetSet($offset, $value)
        {
        }

        #[\ReturnTypeWillChange]
        public function offsetUnset($offset)
        {
        }

        public function __get($key)
        {
            return isset($this->data[$key]) ? $this->data[$key] : null;
        }

        public function __isset($key)
        {
            return isset($this->data[$key]);
        }

        public function __invoke(...$arguments)
        {
            $out = '';

            foreach ($arguments as $arg) {
                if (isset($this->data[$arg])) {
                    $val = $this->data[$arg];
                    if ($val === true) $val = 'true';
                    if ($val === false) $val = 'false';
                    if ($val === null) continue;
                    $out .= ' data-' . $arg . '="' . addslashes($val) . '"';
                }
            }

            return trim($out);
        }
    }

    class PandaRoseUI
    {
        private static $enqueued = false;
        public static $namespace = [];
        public static $top = null;
        public static $deep = 0;
        public static $last = null;
        public static $in = [];
        public static $i = 0;
        public static $test;
        public static $twig;
        public static $tmpl = [];
        public static $repl = [];
        public static $that = [];
        public static $final = [];
        public static $json = null;
        public static $state = [];

        public static $replace = [];
        public static $protect = [];

        public static function replace($find, $replace)
        {
            static::$replace[] = [$find, $replace];
        }

        public static function substitute($string)
        {
            foreach (static::$repl as $find => $replace)
                if (substr_count($string, $find))
                    $string = str_replace($find, $replace, $string);

            return $string;
        }

        public static function protect($path, $dest = null)
        {
            static::$protect[] = [$path, $dest];
        }

        public static function getThis($filepath)
        {
            if (!isset(static::$that[$filepath]))
                static::$that[$filepath] = new PandaRoseUIProxy;
            return static::$that[$filepath];
        }

        public static function addState($state)
        {
            foreach ($state as $k => $v) {
                static::$state[$k] = $v;
            }
        }

        public static function injectState()
        {
            if (count(static::$state)) {
                $json = json_encode(['/' => static::$state]);
                echo <<<HTML
                <script>
                    window.UI_InitialState = JSON.parse('{$json}');
                </script>
                HTML;
            };
        }

        public static function is_ajax()
        {
            if ($_SERVER['CONTENT_TYPE'] === 'application/json')
                return true;
            return false;
        }

        public static function enqueue()
        {
            if (!static::$enqueued) {
                static::$enqueued = true;
                $localJs = __DIR__ . '/build/index.bundle.js';
                $localCss = __DIR__ . '/build/index.bundle.css';
                if (file_exists($localJs)) {
                    wp_enqueue_script('pandarose-ui', plugins_url('build/index.bundle.js', __FILE__));
                } else {
                    // wp_enqueue_script('main', 'https://cdn.jsdelivr.net/gh/pandarose/ui-dist@' . PANDA_UI . '/index.bundle.js');
                    wp_enqueue_script('main', 'http://localhost/build/index.bundle.js');
                }
                if (file_exists($localCss)) {
                    wp_enqueue_style('pandarose-ui', plugins_url('build/index.bundle.css', __FILE__));
                } else {
                    // wp_enqueue_style('main', 'https://cdn.jsdelivr.net/gh/pandarose/ui-dist@' . PANDA_UI . '/index.bundle.css');
                    wp_enqueue_style('main', 'http://localhost/build/index.bundle.css');
                }
            }
        }

        public static function after($callable)
        {
            add_filter(
                'final_output',
                $callable,
                PHP_INT_MAX,
                2
            );
        }

        public static function closeBuffer()
        {
            $out = '';

            $levels = ob_get_level();

            for ($i = 0; $i < $levels; $i++) {
                $out .= ob_get_clean();
            }

            if (!substr_count($_SERVER['REQUEST_URI'], 'wp-login.php')) {
                $uri = trim($_SERVER['REQUEST_URI'], '/');
                foreach (static::$protect as $protect) {
                    list($path, $dest) = $protect;
                    $l = strlen($path);
                    $dest = trim($dest, '/');
                    $path = rtrim($path, '*');
                    $wild = (strlen($path) !== $l);
                    if ((!$wild and ($path === $uri)) or
                        ($wild and substr($uri, 0, strlen($path)) === $path)
                    ) {
                        if (FALSE === is_user_logged_in()) {
                            if ($dest)
                                wp_redirect('/' . $dest);
                            else
                                wp_redirect(wp_login_url(site_url() . $_SERVER['REQUEST_URI']), 302);
                            exit;
                        }
                    }
                }
            }

            return $out;
        }

        public static function encodeWebComponents($buffer)
        {
            while (true) {
                if (
                    !substr_count($buffer, '<!--encode-start-->') or
                    !substr_count($buffer, '<!--encode-end-->')
                ) break;
                $s = strpos($buffer, '<!--encode-start-->');
                $e = strpos($buffer, '<!--encode-end-->');
                $f = $e + strlen('<!--encode-end-->');
                $l = $e - $s;
                $str = substr($buffer, $s, $l);
                $tag = substr($buffer, $s - 14, 14);
                if ($tag !== '<pandarose-ui>') break;
                $pre = substr($buffer, 0, $s - 14);
                $end = substr($buffer, $f, strlen($buffer) - $f);
                $buffer = $pre . $end;
                $uid = bin2hex(random_bytes(16));
                $b64 = base64_encode($str);
                $buffer = $pre .
                    <<<HTML
                <script data-foundation>if(!window.FoundationMarkup) window.FoundationMarkup = {}; window.FoundationMarkup['{$uid}'] = '{$b64}';</script>
                HTML
                    . '<pandarose-ui data-markup="' . $uid . '">'
                    . $end . "\n";
            }
            return $buffer;
        }

        public static function sendBuffer()
        {
            $out = static::closeBuffer();

            if (PandaRoseUI::$json !== null) {
                header('Content-Type: application/json');
                echo json_encode(PandaRoseUI::$json);
                exit;
            }

            $r = count(PandaRoseUI::$repl);

            for ($i = 0; $i < $r; $i++) {
                foreach (PandaRoseUI::$repl as $key => $value) {
                    if (substr_count($out, $key)) {
                        $out = do_shortcode($out);
                        if (!$out) $out = '';
                        if (!$value) $value = '';
                        $out = str_replace($key, $value, $out);
                        unset(PandaRoseUI::$repl[$key]);
                    }
                }
            }

            $scripts = '';

            foreach (PandaRoseUI::$final as $script) {
                $scripts .= '<script>' .
                    'var jq = jQuery;' .
                    'jq(document).ready(function() {' . $script . '});' .
                    '</script>';
            }

            if (substr_count($out, '</body>')) {
                $out = str_replace('</body>', $scripts . '</body>', $out);
            } else {
                $out .= $scripts;
            }

            $out = static::encodeWebComponents($out);

            if (
                count(static::$state) and
                !substr_count($out, 'UI_InitialState') and
                substr_count($out, '</head>')
            ) {
                $json = json_encode(['/' => static::$state]);
                $injectState = <<<HTML
                <script>
                    window.UI_InitialState = JSON.parse('{$json}');
                </script>
                HTML;
                $out = str_replace('</head>', $injectState . '</head>', $out);
            };

            foreach (static::$replace as $replace) {
                $out = str_replace($replace[0], $replace[1], $out);
            }

            echo apply_filters('final_output', $out);
        }

        public static function keyStringGet($array, $keys)
        {
            if (empty($array)) return null;

            $keys = explode('.', $keys);

            do {
                $key = array_shift($keys);

                if (isset($array[$key])) {
                    if (is_array($array[$key]) and !empty($keys)) {
                        $array = $array[$key];
                    } else {
                        return $array[$key];
                    }
                } else {
                    break;
                }
            } while (!empty($keys));

            return null;
        }

        public static function keyString($key, $value, &$out)
        {
            $keys = explode('.', $key);
            $row = &$out;

            for ($i = 0, $end = count($keys) - 1; $i <= $end; $i++) {
                $cur = $keys[$i];
                if (!isset($row[$cur])) {
                    if (isset($keys[$i + 1])) {
                        $row[$cur] = [];
                        $row = &$row[$cur];
                    } else {
                        $row = &$row[$cur];
                    }
                } else {
                    $row = &$row[$cur];
                }
            }

            $row = $value;
        }
    }

    class PandaTwigLoader implements Twig_LoaderInterface
    {
        protected $content;

        public function __construct($content)
        {
            $this->content = $content;
        }
        public function getSourceContext($name)
        {
            if ($name[0] === '#') {
                return new Twig_Source(PandaRoseUI::$tmpl[$name], $name, $name);
            }

            $content = '';

            $content = pandarose($name);

            $repl = md5(random_bytes(32));

            PandaRoseUI::$repl[$repl] = $content;

            return new Twig_Source($repl, $name, $name);
        }

        public function getCacheKey($name)
        {
            return $name;
        }

        public function isFresh($name, $time)
        {
            return true;
        }

        public function exists($name)
        {
            if ($name === '__loaded') {
                return true;
            }

            return true;
        }
    }

    function parseTwig($content, $data = [])
    {
        if (!PandaRoseUI::$twig) {
            $dump = new Twig_Function(
                'dump',
                function ($var) {
                    var_dump($var);
                    return '';
                }
            );
            $slug = new Twig_Function(
                'slug',
                function ($str) {
                    return sanitize_title($str);
                }
            );
            $post = new Twig_Function(
                'post',
                function ($key) {
                    return isset(PandaRoseUI::$in[$key])
                        ? PandaRoseUI::$in[$key] : null;
                }
            );
            $load = new PandaTwigLoader($content);
            PandaRoseUI::$twig = $twig = new Twig_Environment($load, []);
            $twig->addFunction($dump);
            $twig->addFunction($slug);
            $twig->addFunction($post);
            $twig->addExtension(new SwitchTwigExtension());
        } else {
            $twig = PandaRoseUI::$twig;
        }

        PandaRoseUI::$i++;
        PandaRoseUI::$tmpl['#' . PandaRoseUI::$i] = $content;

        $html = $twig->render('#' . PandaRoseUI::$i, $data);

        return $html;
    }

    function foundation_output($lookup = 'default', $data = null, $inner = '')
    {
        return pandarose_output($lookup, $data, $inner);
    }

    function pandarose_output($lookup = 'default', $data = null, $inner = '')
    {
        $output = pandarose($lookup, $data, $inner);

        $output = PandaRoseUI::substitute($output);

        return $output;
    }

    function pandarose($lookup = 'default', $data = null, $inner = '')
    {
        if ($data === null && is_array($lookup)) {
            $data = $lookup;
            $lookup = isset($data['template']) ? $data['template'] : null;
        }

        if (PandaRoseUI::$top === null) {
            PandaRoseUI::$top = $lookup;
        }

        if ($data === null)
            $data = [];

        if (PandaRoseUI::$last === $lookup)
            return '';

        if (isset($lookup) and $lookup) {
            $dir = null;
            if (substr_count($lookup, '/')) {
                list($dir, $template) = explode('/', $lookup, 2);
            } else {
                $template = $lookup;
            }
            if ($dir !== 'core') {
                $scan = [];

                if (!$dir) {
                    foreach (PandaRoseUI::$namespace as $ns) {
                        $scan[] = ['pandarose-' . $ns, $template];
                        if (!substr_count($template, '.')) {
                            $scan[] = ['pandarose-' . $ns, $template . '.yml'];
                            $scan[] = ['pandarose-' . $ns, $template . '.md'];
                            $scan[] = ['pandarose-' . $ns, $template . '.html'];
                        }
                    }
                }
                if (!$dir) {
                    $scan[] = ['pandarose', $template];
                    if (!substr_count($template, '.')) {
                        $scan[] = ['pandarose', $template . '.yml'];
                        $scan[] = ['pandarose', $template . '.md'];
                        $scan[] = ['pandarose', $template . '.html'];
                    }
                } else {
                    $scan[] = ['pandarose-' . $dir, $template];
                    if (!substr_count($template, '.')) {
                        $scan[] = ['pandarose-' . $dir, $template . '.yml'];
                        $scan[] = ['pandarose-' . $dir, $template . '.md'];
                        $scan[] = ['pandarose-' . $dir, $template . '.html'];
                    }
                }
                foreach ($scan as $row) {
                    $post = get_page_by_title($row[1], 'OBJECT', $row[0]);
                    if ($post) {
                        PandaRoseUI::$namespace[] = $dir;
                        PandaRoseUI::$last = $lookup;
                        $out = pandarose_template(null, $post->post_content, $data, $post->post_title, $inner);
                        array_pop(PandaRoseUI::$namespace);
                        PandaRoseUI::$last = null;
                        if (PandaRoseUI::$top === $lookup)
                            PandaRoseUI::$top = null;
                        return $out;
                    }
                }
            } else {
                $lookup = $template;
            }
            if ($template = trim(strtolower($lookup))) {
                $plugin = null;
                if (substr_count($template, '/')) {
                    list($plugin, $template) = explode('/', $template);
                }
                if ($template = (preg_replace('/[^_a-z\d]/i', '', $template))) {
                    $plugin = '';
                    if ($plugin = (preg_replace('/[^_a-z\d]/i', '', $plugin))) {
                        $filepath = __DIR__ . DIRECTORY_SEPARATOR . '..' . DIRECTORY_SEPARATOR . $plugin . DIRECTORY_SEPARATOR . 'pandarose' . DIRECTORY_SEPARATOR . $template . '.php';
                    } else {
                        $filepath = __DIR__ . DIRECTORY_SEPARATOR . 'templates' . DIRECTORY_SEPARATOR . $template . '.php';
                    }
                }
                if (file_exists($filepath)) {
                    PandaRoseUI::$namespace[] = $plugin;
                    PandaRoseUI::$last = $lookup;
                    $out = pandarose_template($filepath, null, $data, null, $inner);
                    array_pop(PandaRoseUI::$namespace);
                    PandaRoseUI::$last = null;
                    if (PandaRoseUI::$top === $lookup)
                        PandaRoseUI::$top = null;
                    return $out;
                }
            }
        }

        return '';
    }

    function pandarose_template(
        $filepath,
        $content = null,
        $data = [],
        $name = null,
        $inner = '',
        $repeating = false
    ) {
        PandaRoseUI::$deep++;

        $verb = $_SERVER['REQUEST_METHOD'];
        if ($repeating) {
            $verb = 'GET';
        }

        top:
        if ($content || ($filepath and file_exists($filepath))) {
            $newHtml = null;
            $in = [];
            if ($verb === 'GET') {
                $in = $_GET;
                $json = null;
            } else if ($json = ($_SERVER['HTTP_ACCEPT'] === 'application/json')) {
                if ($body = file_get_contents('php://input'))
                    if (!is_array($in = json_decode($body, true)))
                        $in = [];
            } else if (isset($_POST['formpost_data'])) {
                if ($body = $_POST['formpost_data'])
                    if (!is_array($in = json_decode($body, true)))
                        $in = [];
                $json = null;
            }
            PandaRoseUI::$in = $in;
            if ($filepath) {
                ob_start();
                $that = PandaRoseUI::getThis($filepath);
                $d = $data;
                if (isset($d['d'])) unset($d['d']);
                $run = function () use ($d, $inner, $filepath) {
                    PandaRoseUI::$i++;

                    $id = 'e' . md5($filepath) . PandaRoseUI::$i;
                    $verb = $_SERVER['REQUEST_METHOD'];

                    if (is_array($d)) {
                        if (!isset($d['id'])) $d['id'] = $id;
                        extract($d);
                    }

                    $data = new PandaRoseData($d);

                    $result = [];

                    if (!$res = require($filepath)) {
                        if ($res !== false)
                            $res = [];
                    }

                    if ($res === false) return false;

                    return is_array($res) ? array_merge($result, $res) : $res;
                };

                $run = $run->bindTo($that, $that);
                $result = $run();

                $output = ob_get_clean();

                if ($result === false) {
                    $output = '';
                    $result = [];
                }
            } else {
                $result = [];
                if ($name) {
                    $extn = pathinfo($name, PATHINFO_EXTENSION);
                    $json = null;
                    switch ($extn) {
                        case 'md':
                            $values = [];
                            $lines = explode("\n", $content);
                            $content = '';
                            foreach ($lines as $line) {
                                if ('set: ' === substr($line, 0, 5)) {
                                    $raw = trim(substr($line, 5));
                                    $val = json_decode($raw, true);
                                    $values = array_merge($values, $val);
                                    continue;
                                }
                                if ('media: ' === substr($line, 0, 7)) {
                                    $src = trim(substr($line, 7));
                                    $attr = [];
                                    $size = 'full';
                                    if (substr_count($src, '{')) {
                                        list($a, $b) = explode('{', $src, 2);
                                        $attr = json_decode('{' . $b, true);
                                        $src = trim($a);
                                    }
                                    if (isset($attr['size'])) {
                                        $size = $attr['size'];
                                        unset($attr['size']);
                                    }
                                    if (!is_numeric($src)) {
                                        $args = array(
                                            'posts_per_page' => 1,
                                            'post_type'      => 'attachment',
                                            'name'           => trim($src),
                                        );
                                        if (
                                            !$attachment = new WP_Query($args)
                                            or !isset($attachment->posts[0])
                                        )
                                            continue;
                                        $src = (int) $attachment->posts[0]->ID;
                                    }
                                    if (is_numeric($src)) {
                                        $line = wp_get_attachment_image((int)$src, $size, false, $attr);
                                    }
                                } else if ('include: ' === substr($line, 0, 9)) {
                                    $incl = trim(substr($line, 9));
                                    $repl = md5(random_bytes(32));
                                    $data = array_merge($values, []);
                                    if (substr_count($incl, '{')) {
                                        list($a, $b) = explode('{', $incl, 2);
                                        $data = array_merge(
                                            $data,
                                            json_decode('{' . $b, true)
                                        );
                                        $incl = trim($a);
                                    }
                                    PandaRoseUI::$repl[$repl] = pandarose($incl, $data);
                                    $line = $repl;
                                }
                                $content .= $line . "\n";
                            }
                            $parsedown = new Parsedown();
                            $output = $parsedown->text($content);
                            foreach ($values as $k => $v)
                                $output = str_replace(
                                    '%%' . strtoupper($k) . '%%',
                                    $v,
                                    $output
                                );
                            break;
                        case 'json':
                            if (!$json = json_decode($content, true)) {
                                $json = [];
                            }
                        case 'yml':
                        case 'yaml':
                            $result = ($json !== null)
                                ? $json : Yaml::parse($content);
                            if (isset($result['render'])) {
                                $render = $result['render'];
                                unset($result['render']);
                                PandaRoseUI::$test = 1;
                                $output = pandarose($render, $result);
                            }
                            break;
                        case 'scss':
                            $scss = new Compiler();
                            $scss->setVariables([]);
                            $content = $scss->compile($content);
                        case 'css':
                            $output = '<style>' . $content . '</style>';
                            break;
                        default:
                            $output = parseTwig($content, $data);
                    }
                } else {
                    $output = parseTwig($content, $data);
                }
            }

            if (!$output) $output = '';

            $output = do_shortcode($output);

            if ($json)
                ob_clean();

            $responders = [];
            $v = ['state' => []];
            $raw = false;
            $script = '';
            if (is_iterable($result)) {
                foreach ($result as $q => $r) {
                    if ($q === 'raw' and $r === true)
                        $raw = true;
                    else if ($q === 'script')
                        $script = $r;
                    else if ($q === 'final')
                        PandaRoseUI::$final[$filepath] = $r;
                    else if ($q === 'jquery') {
                        $raw = true;
                        PandaRoseUI::$final[$filepath] = $r;
                    } else if (is_string($q) and substr($q, 0, 5) === 'state') {
                        if ($q === 'state') {
                            $v['state'] = array_merge(
                                $v['state'],
                                $r
                            );
                        } else
                            PandaRoseUI::keyString($q, $r, $v);
                    }

                    if (!$repeating)
                        $responders[$q] = $r;
                }
                if ($result instanceof Generator) {
                    if ($ret = $result->getReturn())
                        if (!$repeating)
                            $responders[$q] = $r;
                }
            }

            if ($raw and PandaRoseUI::$deep > 1) return '';

            if (is_callable($result))
                $responders[] = $result;
            foreach ($responders as $key => $result) {
                $keyVerb = 'POST';
                if (!is_string($key)) $key = '';
                if (substr_count($key, '?')) {
                    list($keyVerb, $key) = explode('?', $key);
                }
                if (strtoupper($key) === $key) {
                    $keyVerb = $key;
                    $key = '';
                }
                parse_str($key, $bits);
                if (count($bits)) {
                    foreach ($bits as $bit => $val) {
                        if ($val === '' && isset($in[$bit])) {
                        } else if ($val && isset($in[$bit]) and $in[$bit] == $val) {
                        } else {
                            continue 2;
                        }
                    }
                }
                $reflect = new ReflectionFunction($result);
                $args = [];
                foreach ($reflect->getParameters() as $param) {
                    $name = $param->getName();
                    if ($name === 'in') {
                        $args[] = $in;
                    } else {
                        if (isset($in[$name]))
                            $args[] = $in[$name];
                        else if ($param->allowsNull())
                            $args[] = null;
                    }
                }
                if ($verb !== $keyVerb) continue;
                try {
                    $result = $result(...$args);
                    if ($verb !== 'GET') {
                        if (is_iterable($result)) {
                            again:
                            foreach ($result as $q => $r) {
                                if (is_string($q) and substr($q, 0, 5) === 'state') {
                                    PandaRoseUI::keyString($q, $r, $v);
                                } else if (is_string($q)) {
                                    $q = 'state.' . $q;
                                    PandaRoseUI::keyString($q, $r, $v);
                                }
                            }
                            if ($result instanceof Generator) {
                                if ($ret = $result->getReturn()) {
                                    $result = $ret;
                                    if (is_iterable($ret))
                                        goto again;
                                }
                            }
                        }
                        $page = [
                            'data' => count($v['state']) ? $v['state'] : new stdClass
                        ];
                        if (is_string($result))
                            $page['__html'] = base64_encode($result);
                        else if ($result === true) {
                            if (!PandaRoseUI::is_ajax()) {
                                array_pop(PandaRoseUI::$namespace);
                                PandaRoseUI::$last = null;
                                $ns = array_pop(PandaRoseUI::$namespace);
                                $res = pandarose_template(
                                    $filepath,
                                    $content,
                                    $data,
                                    $name,
                                    $inner,
                                    true
                                );
                                PandaRoseUI::$namespace[] = $ns;
                                return $res;
                            }
                            PandaRoseUI::$json = [
                                '__signals' => [
                                    ['571c00301174ce0e54103bb043bd64c1', new stdClass]
                                ]
                            ];
                        } else {
                            PandaRoseUI::$json = [
                                'page' => $page
                            ];
                        }
                        exit;
                        // header('Content-Type: application/json');
                        // echo json_encode([
                        //     'page' => $page
                        // ]);
                        // exit;
                    } else {
                        if ($result === true) {
                            $uri = '/' . trim($_SERVER['REQUEST_URI'], '/');
                            list($uri) = explode('?', $uri);
                            header('Location: ' . $uri);
                            exit;
                        }
                    }
                } catch (\Throwable $e) {
                }
            }

            $html = $newHtml ?? $output;

            if ($json) {
                PandaRoseUI::$json = [
                    'page' => [
                        'html' => $html
                    ]
                ];
                exit;
                // header('Content-Type: application/json');
                // echo json_encode([
                //     'page' => [
                //         'html' => $html
                //     ]
                // ]);
                // exit;
            }

            if ($verb === 'GET' or $repeating) {
                if (
                    !$raw and
                    !substr_count($html, '<pandarose-ui') and
                    PandaRoseUI::$deep < 2
                ) {
                    $html = <<<HTML
                <pandarose-ui><!--encode-start-->{$html}<!--encode-end--></pandarose-ui>
                HTML;
                }

                if ($raw and $script) {
                    $html .= '<script>' .
                        'var jq = jQuery;' .
                        'jq(document).ready(function() {' . $script . '});' .
                        '</script>';
                }

                PandaRoseUI::$deep--;

                if (count($v['state']))
                    PandaRoseUI::addState($v['state']);

                return $html;
            }
        }

        PandaRoseUI::$deep--;
    }

    function pandarose_ui_shortcode($args, $content = '')
    {
        wp_enqueue_script("jquery");

        PandaRoseUI::enqueue();

        if (!is_array($args)) $args = [];

        return do_shortcode(pandarose($args, null, $content));
    }

    function pandarose_ui_admin()
    {
        add_menu_page('Foundation UI', 'Foundation UI', 'manage_options', 'ui', 'admin_page', 'dashicons-block-default');
    }

    function admin_page()
    {
        PandaRoseUI::enqueue();
        $admin = isset($_GET['admin']) ? $_GET['admin'] : 'default';
        if ($admin = (preg_replace('/[^_a-z\d]/i', '', $admin))) {
            $filepath = __DIR__ . DIRECTORY_SEPARATOR . 'admin' . DIRECTORY_SEPARATOR . $admin . '.php';
            echo <<<JS
        <script>
            window.PushHandler = function(path) {
                path = window.FoundationHelpers.lodash.trim(path, '/');
                window.FoundationHelpers.insertUrlParam('admin', path, true);
            }
        </script>
        JS;
            if ($result = pandarose_template($filepath)) {
                echo $result;
            } else {
                echo 'not found';
            }
        } else {
            echo 'Not Found';
        }
    }

    function pandarose_response_post_type($name = null, $opts = [])
    {
        if (is_array($name)) {
            $opts = $name;
            $name = null;
        }

        $suffix = 'response';

        if ($name) {
            $suffix .= '-' . $name;
            if (!isset($opts['name']))
                $opts['name'] = ucfirst($name);
        } else if (!$name) {
            $opts['top'] = true;
        }

        return pandarose_post_type($suffix, $opts);
    }

    function pandarose_post_type($suffix = '', $opts = [])
    {
        if (is_array($suffix)) {
            $opts = $suffix;
            $suffix = '';
        }

        $templates = true;
        $prefix = 'pandarose';
        $tree = 'Templates';

        if ('response-' === substr($suffix, 0, 9)) {
            $suffix = substr($suffix, 9);
            $prefix = 'response';
            $tree = 'Responses';
            $templates = false;
        } else if ($suffix === 'response') {
            $prefix = 'response';
            $suffix = '';
            $tree = 'Responses';
            $templates = false;
        }

        if (is_string($opts))
            $opts = ['name' => $opts];

        if ($suffix and (!isset($opts['top']) or !$opts['top'])) {
            $opts['show_in_menu'] = 'edit.php?post_type=' . $prefix;
            if (!isset($opts['name']))
                $opts['name'] = ucfirst(strtolower($suffix));
            if (isset($opts['top']))
                unset($opts['top']);
        }

        if (!isset($opts['name']))
            $opts['name'] = $tree;

        $labels = [
            'name'          => $opts['name'],
            'singular_name' => isset($opts['singular_name'])
                ? $opts['singular_name'] : $opts['name'],
            'menu_name'     => isset($opts['menu_name'])
                ? $opts['menu_name'] : $opts['name'],
        ];

        if (isset($opts['singular_name'])) unset($opts['singular_name']);
        if (isset($opts['menu_name'])) unset($opts['menu_name']);

        unset($opts['name']);

        $orig = $opts;
        $opts = array_merge([
            'labels'             => $labels,
            'public'             => false,
            'publicly_queryable' => false,
            'show_ui'            => true,
            'show_in_menu'       => true,
            'query_var'          => true,
            'capability_type'    => 'post',
            'taxonomies'         => array('category'),
            'has_archive'        => true,
            'hierarchical'       => false,
            'menu_position'      => null,
            'supports'           => ['title', 'editor', 'custom-fields', 'revisions', 'comments', 'thumbnail'],
        ], $opts);

        if (!$templates and !isset($orig['capabilities'])) {
            $opts['capabilities'] = [
                'create_posts' => 'do_not_allow',
                'edit_post' => true,
            ];
            $opts['map_meta_cap'] = true;
        }

        register_post_type($suffix
            ? $prefix . '-' . $suffix
            : $prefix, $opts);
    }

    function pandarose_ui_input_shortcode($args)
    {
        $parts = ['input'];

        if (isset($args['type'])) {
            $parts['type'] = $args['type'];
            switch ($args['type']) {
                case 'textarea':
                    $parts = ['textarea'];
                    break;
                case 'checkbox':
                case 'radio':
                    if (!isset($args['label']) and isset($args['value']))
                        $args['label'] = $args['value'];
                    break;
            }
        } else {
            $parts['type'] = 'text';
        }

        if (isset($args['label']) and !isset($args['name'])) {
            $args['name'] = sanitize_title($args['label']);
            if (isset($args['prefix']) and $args['prefix']) {
                $args['name'] = $args['prefix'] . $args['name'];
                unset($args['prefix']);
            }
        }

        foreach ($args as $k => $v) {
            if (is_numeric($k)) {
                if (substr($v, 0, 3) === 'is-') {
                    $k = substr($v, 3);
                    $v = true;
                } elseif (substr($v, 0, 4) === 'not-') {
                    $k = substr($v, 4);
                    $v = false;
                } else {
                    continue;
                }
            }
            $parts[$k] = $v;
        }

        $out = '';

        foreach ($parts as $k => $v) {
            if (is_numeric($k)) {
                $out .= $v . ' ';
            } else {
                $out .= $k . '="' . addslashes($v) . '" ';
            }
        }

        $out = '<' . trim($out) . '>';

        if ($parts[0] === 'textarea')
            $out .= '</textarea>';

        return $out;
    }

    function pandarose_inject_editor($content, $opts = [])
    {
        $opts = array_merge([
            'textLabel' => 'Code'
        ], $opts);

        PandaRoseUI::after(
            function ($out) use ($content, $opts) {
                if (!strpos($out, '<div id="wp-content-editor-container"'))
                    return $out;
                $out = <<<HTML
                <script>
                    jQuery('#wp-content-editor-container').hide();
                    jQuery('#wp-content-editor-tools').hide();
                    jQuery('#content-tmce').hide();
                </script>
                <style>
                    #wp-content-editor-container { 
                        display: none !important;
                    }
                    #content-tmce {
                        display: none !important;
                    }
                    #content-html {
                        opacity: 0;
                    }
                    #wp-word-count {
                        display: none !important;
                    }
                </style>
                HTML . $out;
                $pos = strpos($out, '<div id="wp-content-editor-container"');
                if (
                    '<pandarose-ui' === substr($content, 0, 13) or
                    '<script data-foundation' === substr($content, 0, 23)
                ) {
                    $replace = $content;
                } else {
                    $replace = '<pandarose-ui>' . $content . '</pandarose-ui>';
                }
                $out = substr($out, 0, $pos)
                    . $replace
                    . substr($out, $pos, strlen($out) - strlen($pos));
                $out .= <<<HTML
            <script>
                jQuery('#wp-content-editor-tools').hide();
                jQuery('#wp-content-wrap').css('margin-top', '-50px');
            </script>
            HTML;
                return $out;
            }
        );
    }

    function pandarose_monaco($language = 'twig', $opts = [])
    {
        pandarose_inject_editor(
            <<<HTML
        <div id="pandarose-editor" style="opacity: 0; min-height: 500px">
            <component type="monaco" language="$language" bind="code" expr="typeof bound.value !== 'undefined'" width="100%" height="500">
            </component>
        </div>
        <script code-bind="code">
            switch(changed) {
                case 'code':
                    if(self.started)
                        jQuery('textarea.wp-editor-area').val(bound.code);
                    self.started = true;
                    break;
                default:
                    var code = jQuery('textarea.wp-editor-area').val();
                    mutate.code.change(code || '');
                    jQuery('#pandarose-editor').css('opacity', 1);
            }
        </script>
        HTML,
            $opts
        );
    }

    add_filter('wp_default_editor', function ($editor) {
        $type = get_post_type();
        if ($type === 'pandarose' or substr($type, 0, 10) === 'pandarose-') {
            $language = 'twig';
            if ($post = get_post()) {
                $title = $post->post_title;
                $extn = pathinfo($title, PATHINFO_EXTENSION);
                $opts = [];
                switch ($extn) {
                    case 'yml':
                    case 'yaml':
                        $language = 'yaml';
                        break;
                    case 'js':
                    case 'jsx':
                        $language = 'javascript';
                        break;
                    case 'md':
                        $language = 'markdown';
                        break;
                    case 'css':
                        $language = 'css';
                        break;
                    case 'scss':
                        $language = 'scss';
                        break;
                    case 'json':
                        $language = 'json';
                        break;
                }
            }
            pandarose_monaco($language, $opts);
            return 'html';
        } else if ($type === 'response' or substr($type, 0, 10) === 'response-') {
            $post = get_post();
            $content = json_decode(base64_decode($post->post_content), true);
            $editor = pandarose_output('steps', $content);
            $editor = PandaRoseUI::encodeWebComponents($editor);
            pandarose_inject_editor(
                <<<HTML
            $editor
            HTML,
                []
            );

            return 'html';
        }
        return 'tinymce';
    });

    add_action('shutdown', [PandaRoseUI::class, 'sendBuffer'], 0);
    add_action('admin_menu', 'pandarose_ui_admin');
    add_action('admin_enqueue_scripts', [PandaRoseUI::class, 'enqueue']);
    add_shortcode('pandarose_ui', 'pandarose_ui_shortcode');
    add_shortcode('pandarose-ui', 'pandarose_ui_shortcode');
    add_shortcode('pandarose', 'pandarose_ui_shortcode');
    add_shortcode('ui', 'pandarose_ui_shortcode');
    add_shortcode('ui_input', 'pandarose_ui_input_shortcode');
    add_shortcode('ui-input', 'pandarose_ui_input_shortcode');
    add_action('wp_head', [PandaRoseUI::class, 'injectState']);

    add_action(
        'init',
        function () {
            pandarose_post_type(['menu_icon' => 'dashicons-media-code']);
            pandarose_post_type('form');
            pandarose_response_post_type(['menu_icon' => 'dashicons-media-text']);
        }
    );

    add_action(
        'admin_init',
        function () {
            pandarose_post_type();
            pandarose_post_type('form');
            pandarose_response_post_type();
        }
    );

    function foundation_template_type()
    {
    }

    function foundation_response_type()
    {
    }

    function foundation_object_type()
    {
    }

    class_alias('PandaRoseUI', 'FoundationUI');

    ob_start();

    add_action('add_meta_boxes', function () {

        // add_meta_box(
        //     'global-notice',
        //     __( 'Global Notice', 'sitepoint' ),
        //     function() {
        //         return 'test';
        //     },
        //     'response'
        // );

    });
}
